// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 12:35 BRST
// Last edited on December 22 of 2017, at 12:11 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/arch/port.h>
#include <chicago/console.h>

PUInt16 ConsoleMemory = (PUInt16)0xB8000;
UInt16 ConsoleCursorX = 0;
UInt16 ConsoleCursorY = 0;
UInt16 ConsoleAttribute = 0x07;

Void ConsoleUpdateCursor(Void)
{
	UInt16 pos = (UInt16)((ConsoleCursorY * 80) + ConsoleCursorX);		// Get real cursor position
	PortOutByte(0x3D4, 0x0F);		// Cursor low port to VGA index register
	PortOutByte(0x3D5, (UInt8)(pos & 0xFF));
	PortOutByte(0x3D4, 0x0E);		// Cursor high port to VGA index register
	PortOutByte(0x3D5, (UInt8)((pos >> 8) & 0xFF));
}

Void ConsoleScroll(Void)
{
	if (ConsoleCursorY >= 25)		// Only scroll if cursor is out of screen
	{
		for (UInt16 i = 0; i < 80 * 24; i++)
			ConsoleMemory[i] = ConsoleMemory[i + 80];		// Move one line up
		for (UInt16 i = 24 * 80; i < 25 * 80; i++)
			ConsoleMemory[i] = ConsoleAttribute << 8 | ' ';		// Just put an space
		ConsoleCursorY = 24;		// Set y to last line
		ConsoleUpdateCursor();		// Update cursor
	}
}

UInt8 ConsoleGetBackground(Void)
{
	return (ConsoleAttribute & 0xF0) >> 4;		// Extract high nibble
}

UInt8 ConsoleGetForeground(Void)
{
	return ConsoleAttribute & 0x0F;		// Extract low nibble
}

Void ConsoleSetBackground(UInt8 color)
{
	ConsoleAttribute = (color & 0xF0) << 4 | ConsoleGetForeground() & 0x0F;
}

Void ConsoleSetForeground(UInt8 color)
{
	ConsoleAttribute = (ConsoleGetBackground() & 0x0F) << 4 | color & 0x0F;
}

Void ConsoleSetColor(UInt8 back, UInt8 fore)
{
	ConsoleAttribute = (back & 0x0F) << 4 | fore & 0x0F;
}

Void ConsoleClear(Void)
{
	for (UInt16 i = 0; i < 80 * 25; i++)
		ConsoleMemory[i] = ConsoleAttribute << 8 | ' ';		// Just put an space
	ConsoleCursorX = ConsoleCursorY = 0;		// Set x and y to 0
	ConsoleUpdateCursor();		// Update cursor
}

Void ConsoleWriteCharacter(Char data)
{
	switch (data)
	{
		case '\b':		// Backspace
		{
			if (ConsoleCursorY > 0)		// Are we in line 0?
			{
				if (ConsoleCursorX > 0)		// No, so is cursor x greater than 0?
					ConsoleCursorX--;		// Yes, so just decrease it
				else
				{
					ConsoleCursorY--;		// No, so decrease y
					ConsoleCursorX = 79;		// And set x to 79
				}
			}
			else if (ConsoleCursorX > 0)		// Yes, so decrease x if it's greater than 0
				ConsoleCursorX--;
			break;
		}
		case '\n':		// New line
		{
			ConsoleCursorY++;		// Increase y
			ConsoleCursorX = 0;		// And set x to zero
			break;
		}
		case '\r':		// Carriage return
		{
			ConsoleCursorX = 0;		// Just set x to zero
			break;
		}
		case '\t':
		{
			ConsoleCursorX += 4;		// Add 4 to x
			break;
		}
		default:		// Character
		{
			UInt16 pos = (UInt16)((ConsoleCursorY * 80) + ConsoleCursorX);		// Get real cursor position
			ConsoleMemory[pos] = ConsoleAttribute << 8 | data;		// Put character in screen
			ConsoleCursorX++;		// Increase x
		}
	}

	if (ConsoleCursorX >= 80)		// Is x out of current line?
	{
		ConsoleCursorY++;		// Yes, so increase y
		ConsoleCursorX = 0;		// And set x to zero
	}

	ConsoleScroll();		// Scroll if necessary
	ConsoleUpdateCursor();		// Update cursor
}

Void ConsoleWriteString(PChar data)
{
	for (Int i = 0; data[i]; i++)
		ConsoleWriteCharacter(data[i]);
}

Void ConsoleWriteInteger(ULong data, UInt8 base)
{
	ULong n = data / base;
	Long r = data % base;

	if (r < 0)
	{
		r += base;
		n--;
	}

	if (data >= base)
		ConsoleWriteInteger(n, base);

	ConsoleWriteCharacter("0123456789ABCDEF"[r]);
}

Void ConsoleWriteFormated(PChar data, ...)
{
	VariadicList va;
	VariadicStart(va, data);		// Lets start our va list with the arguments provided by the user (if any)

	for (Int i = 0; data[i]; i++)
	{
		if (data[i] != '%')		// It's an % (integer, string, character or other)?
			ConsoleWriteCharacter(data[i]);		// No, so just display it
		else
		{
			switch (data[++i])		// Yes, so lets parse it!
			{
				case 's':		// String
				{
					ConsoleWriteString((PChar)VariadicArg(va, PChar));
					break;
				}
				case 'c':		// Character
				{
					ConsoleWriteCharacter((Char)VariadicArg(va, Int));
					break;
				}
				case 'd':		// Integer (Decimal)
				{
					ConsoleWriteInteger((ULong)VariadicArg(va, ULong), 10);
					break;
				}
				case 'x':		// Integer (Hexadecimal)
				{
					ConsoleWriteInteger((ULong)VariadicArg(va, ULong), 16);
					break;
				}
				default:		// None of other, so just display it
				{
					ConsoleWriteCharacter(data[i]);
					break;
				}
			}
		}
	}

	VariadicEnd(va);
}

Void Message(Int level, PChar data, ...)
{
	UInt16 old = ConsoleGetForeground();
	
	if (level == MESSAGE_LEVEL_INFO)
		ConsoleSetForeground(CONSOLE_COLOR_GREEN);
	else if (level == MESSAGE_LEVEL_WARNING)
		ConsoleSetForeground(CONSOLE_COLOR_YELLOW);
	else if (level == MESSAGE_LEVEL_ERROR)
		ConsoleSetForeground(CONSOLE_COLOR_RED);
	
	VariadicList va;
	VariadicStart(va, data);		// Lets start our va list with the arguments provided by the user (if any)

	for (Int i = 0; data[i]; i++)
	{
		if (data[i] != '%')		// It's an % (integer, string, character or other)?
			ConsoleWriteCharacter(data[i]);		// No, so just display it
		else
		{
			switch (data[++i])		// Yes, so lets parse it!
			{
				case 's':		// String
				{
					ConsoleWriteString((PChar)VariadicArg(va, PChar));
					break;
				}
				case 'c':		// Character
				{
					ConsoleWriteCharacter((Char)VariadicArg(va, Int));
					break;
				}
				case 'd':		// Integer (Decimal)
				{
					ConsoleWriteInteger((ULong)VariadicArg(va, ULong), 10);
					break;
				}
				case 'x':		// Integer (Hexadecimal)
				{
					ConsoleWriteInteger((ULong)VariadicArg(va, ULong), 16);
					break;
				}
				default:		// None of other, so just display iy
				{
					ConsoleWriteCharacter(data[i]);
					break;
				}
			}
		}
	}

	VariadicEnd(va);
	
	ConsoleSetForeground(old);
}

Void Panic(PChar data, ...)
{
	ConsoleSetColor(CONSOLE_COLOR_RED, CONSOLE_COLOR_WHITE);
	//ConsoleClear();
	
	VariadicList va;
	VariadicStart(va, data);		// Lets start our va list with the arguments provided by the user (if any)

	for (Int i = 0; data[i]; i++)
	{
		if (data[i] != '%')		// It's an % (integer, string, character or other)?
			ConsoleWriteCharacter(data[i]);		// No, so just display it
		else
		{
			switch (data[++i])		// Yes, so lets parse it!
			{
				case 's':		// String
				{
					ConsoleWriteString((PChar)VariadicArg(va, PChar));
					break;
				}
				case 'c':		// Character
				{
					ConsoleWriteCharacter((Char)VariadicArg(va, Int));
					break;
				}
				case 'd':		// Integer (Decimal)
				{
					ConsoleWriteInteger((ULong)VariadicArg(va, ULong), 10);
					break;
				}
				case 'x':		// Integer (Hexadecimal)
				{
					ConsoleWriteInteger((ULong)VariadicArg(va, ULong), 16);
					break;
				}
				default:		// None of other, so just display it
				{
					ConsoleWriteCharacter(data[i]);
					break;
				}
			}
		}
	}

	VariadicEnd(va);
	
	while (1)
		Asm("hlt");
}
