// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 13:56 BRST
// Last edited on December 18 of 2017, at 21:11 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/string.h>
#include <chicago/console.h>
#include <chicago/alloc.h>
#include <chicago/arch/paging.h>

extern UIntPtr KernelEnd;

UIntPtr PagingAllocPointer = (UIntPtr)&KernelEnd;
PUInt32 PagingFrames = Null;
UInt32 PagingFrameCount = 0;
PPageDirectory PagingCurrentDirectory = Null;
PPageDirectory PagingKernelDirectory = Null;

UIntPtr PagingRealAlloc(UInt32 size, Boolean align, PUIntPtr phys)
{
	if (PagingAllocPointer == (UIntPtr)-1)
	{
		UIntPtr addr;
		
		if (align)
			addr = MemoryAlignedAlloc(size, 0x1000);
		else
			addr = MemoryAlloc(size);
		
		if (phys)
		{
			PPage page = PagingGetPage(addr, 0, PagingKernelDirectory);
			*phys = page -> frame * 0x1000 + (addr & 0xFFF);
		}
		
		return addr;
	}
	
	if (align == True && (PagingAllocPointer & 0xFFFFF000))		// Align the address to page size?
	{
		PagingAllocPointer &= 0xFFFFF000;		// Yes
		PagingAllocPointer += 0x1000;
	}
	
	if (phys)		// Return the physical address?
		*phys = PagingAllocPointer;		// Yes
	
	UIntPtr ret = PagingAllocPointer;
	
	PagingAllocPointer += size;
	
	return ret;
}

UIntPtr PagingAlloc(UInt32 size)
{
	return PagingRealAlloc(size, 0, Null);		// Vanilla alloc, without align and without returning the physical address
}

UIntPtr PagingAlignedAlloc(UInt32 size)
{
	return PagingRealAlloc(size, 1, Null);		// Aligned alloc
}

UIntPtr PagingAllocP(UInt32 size, PUIntPtr phys)
{
	return PagingRealAlloc(size, 0, phys);		// Also return the physical address
}

UIntPtr PagingAlignedAllocP(UInt32 size, PUIntPtr phys)
{
	return PagingRealAlloc(size, 1, phys);		// Aligned alloc and returns the physical address
}

static Void PagingSetFrame(UInt32 addr)
{
	UInt32 frame = addr / 0x1000;
	UInt32 idx = frame / 32;
	UInt32 off = frame % 32;
	
	PagingFrames[idx] |= (1 << off);
}

static Void PagingClearFrame(UInt32 addr)
{
	UInt32 frame = addr / 0x1000;
	UInt32 idx = frame / 32;
	UInt32 off = frame % 32;
	
	PagingFrames[idx] &= ~(1 << off);
}

static UInt32 PagingFirstFreeFrame(Void)
{
	for (UInt32 i = 0; i < (PagingFrameCount / 32); i++)
	{
		if (PagingFrames[i] != 0xFFFFFFFF)		// Any free bit?
		{
			for (UInt32 j = 0; j < 32; j++)		// Yes, let's find it and return it
			{
				if (!(PagingFrames[i] & (1 << j)))
					return i * 4 * 8 + j;
			}
		}
	}
	
	return (UInt32)-1;
}

Void PagingAllocFrame(PPage page, Int krnl, Int write)
{
	if (page -> frame)		// Frame already allocated?
		return;		// Yes, so just return
	else
	{
		UInt32 idx = PagingFirstFreeFrame();
		
		if (idx == (UInt32)-1)		// If idx is -1 then we don't have any free frame... this is an problem
			Panic("Sorry, CHIcago got an fatal error and halted.\nThis fault happened because we run out of free frames.");
		
		PagingSetFrame(idx * 0x1000);		// THIS FRAME BELONG TO US
		
		page -> present = 1;		// Mark it as present
		page -> rw = write ? 1 : 0;		// Writeable?
		page -> user = krnl ? 0 : 1;		// User mode?
		page -> frame = idx;
	}
}

Void PagingFreeFrame(PPage page)
{
	UInt32 frame;
	
	if (!(frame = page -> frame))		// We have any frame to free?
		return;		// No, so just return
	else
	{
		PagingClearFrame(frame);		// Free the frame
		page -> frame = 0;		// Clear the frame in the page
	}
}

Void PagingEnable(Void)
{
	UInt32 cr0;
	Asm Volatile("mov %%cr0, %0" : "=r"(cr0));		// Get the cr0
	cr0 |= 0x80000000;		// Set the paging bit
	Asm Volatile("mov %0, %%cr0" :: "r"(cr0));		// Write it back
}

Void PagingSwitchDirectory(PPageDirectory dir)
{
	PagingCurrentDirectory = dir;
	Asm Volatile("mov %0, %%cr3" :: "r"(dir -> physical_address));
}

PPage PagingGetPage(UInt32 addr, Int make, PPageDirectory dir)
{
	UInt32 idx = addr / 0x1000;		// Turn the address into an index
	UInt32 tidx = idx / 1024;		// Find the page table containing this address
	
	if (dir -> tables[tidx])		// Table already signed?
		return &dir -> tables[tidx] -> pages[idx % 1024];		// Yes, so return it
	else if (make)		// Make it?
	{
		UIntPtr tmp;		// Yes!
		
		dir -> tables[tidx] = (PPageTable)PagingAlignedAllocP(sizeof(PageTable), &tmp);
		
		MemorySet((PUInt8)dir -> tables[tidx], 0, 0x1000);
		
		dir -> physical_tables[tidx] = tmp | 7;
		
		return &dir -> tables[tidx] -> pages[idx % 1024];
	}
	else
		return 0;		// No...
}

PPageTable PagingCloneTable(PPageTable src, PUInt32 phys)
{
	PPageTable table = (PPageTable)PagingAlignedAllocP(sizeof(PageTable), phys);		// Page tables should be page aligned
	
	MemorySet((PUInt8)table, 0, sizeof(PageTable));		// Zero it
	
	for (Int i = 0; i < 1024; i++)
	{
		if (!src -> pages[i].frame)
			continue;
		
		PagingAllocFrame(&table -> pages[i], 0, 0);		// Alloc a new frame
		
		if (src -> pages[i].present)		// Copy the present bit
			table -> pages[i].present = 1;
		if (src -> pages[i].rw)		// Copy the rw bit
			table -> pages[i].rw = 1;
		if (src -> pages[i].user)		// Copy the user bit
			table -> pages[i].user = 1;
		if (src -> pages[i].accessed)		// Copy the accessed bit
			table -> pages[i].accessed = 1;
		if (src -> pages[i].dirty)		// Copy the dirty bit
			table -> pages[i].dirty = 1;
		
		PagingCopyPhysicalPage(src -> pages[i].frame * 0x1000, table -> pages[i].frame * 0x1000);		// Now, physically copy the data
	}
	
	return table;
}

PPageDirectory PagingCloneDirectory(PPageDirectory src)
{
	UInt32 phys;
	PPageDirectory dir = (PPageDirectory)PagingAlignedAllocP(sizeof(PageDirectory), &phys);		// Page directorys should be page aligned
	
	MemorySet((PUInt8)dir, 0, sizeof(PageDirectory));		// Zero it
	
	UInt32 off = (UInt32)dir -> physical_tables - (UInt32)dir;
	
	dir -> physical_address = phys + off;		// Set the physical address of dir -> physical_tables
	
	for (Int i = 0; i < 1024; i++)
	{
		if (!src -> tables[i] || (UIntPtr)src -> tables[i] == (UIntPtr)0xFFFFFFFF)
			continue;
		
		if (PagingKernelDirectory -> tables[i] == src -> tables[i])
		{
			dir -> tables[i] = src -> tables[i];		// Copy from the kernel
			dir -> physical_tables[i] = src -> physical_tables[i];
		}
		else
		{
			UInt32 phys;		// Copy from src
			dir -> tables[i] = PagingCloneTable(src -> tables[i], &phys);
			dir -> physical_tables[i] = phys | 0x07;
		}
	}
	
	return dir;
}

Void PagingInit(PMultibootHeader mbheader)
{
	/* First, init the alloc pointer (put it after all modules) */
	
	if ((mbheader -> flags && (1 << 3)) && (mbheader -> mods_count > 0))		// Modules avaliable?
	{
		UIntPtr lastmod = KernelEnd;		// Yes, so let's put our alloc pointer after all modules
		
		for (Int i = 0; i < mbheader -> mods_count; i++)
		{
			PMultibootModule mod = &((PMultibootModule)mbheader -> mods_address)[i];
			UIntPtr mend = mod -> mod_end;
			
			if ((UIntPtr)mod + sizeof(MultibootModule) > lastmod)
				lastmod = (UIntPtr)mod + sizeof(MultibootModule);
			if (lastmod < mend)
				lastmod = mend;
		}
		
		PagingAllocPointer = lastmod;
	}
	
	/* Now, init the bitset */
	
	PagingFrameCount = (mbheader -> mem_lower + mbheader -> mem_upper) / 4;
	PagingFrames = (PUInt32)PagingAlloc(PagingFrameCount / 32);
	
	MemorySet((PUInt8)PagingFrames, 0, (PagingFrameCount / 32));
	
	/* Init the kernel directory */
	
	PagingKernelDirectory = (PPageDirectory)PagingAlignedAlloc(sizeof(PageDirectory));
	
	MemorySet((PUInt8)PagingKernelDirectory, 0, sizeof(PageDirectory));
	
	PagingKernelDirectory -> physical_address = (UInt32)PagingKernelDirectory -> physical_tables;
	
	Int i = 0;
	while (i < PagingAllocPointer)
	{
		PagingAllocFrame(PagingGetPage(i, 1, PagingKernelDirectory), 0, 0);		// Kernel code is readable from userspace, but isn't writeable
		i += 0x1000;
	}
	
	/* Switch page directory and enable paging */
	
	PagingCurrentDirectory = PagingCloneDirectory(PagingKernelDirectory);
	
	PagingSwitchDirectory(PagingKernelDirectory);
	PagingEnable();
}
