// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 12:09 BRST
// Last edited on March 05 of 2018, at 12:06 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/version.h>
#include <chicago/multiboot.h>
#include <chicago/string.h>
#include <chicago/console.h>
#include <chicago/arch.h>
#include <chicago/arch/port.h>
#include <chicago/arch/idt.h>
#include <chicago/arch/paging.h>
#include <chicago/timer.h>
#include <chicago/keyboard.h>
#include <chicago/task.h>
#include <chicago/alloc.h>
#include <chicago/list.h>

/* Others */

PChar StringDuplicate(PChar src)
{
	PChar ret = (PChar)MemoryAlloc(StringGetLength(src) + 1);
	
	if (ret)
		StringCopy(ret, src);
	
	return ret;
}

/* FS */

#define FS_FLAG_FILE 1
#define FS_FLAG_DIR 2

typedef struct FSFileStruct
{
	Char name[256];
	UInt32 flags;
	UInt32 mask;
	UInt32 length;
	UIntPtr offset;
	UIntPtr (*read)(struct FSFileStruct*, UIntPtr, PChar, UIntPtr);		// Returns UIntPtr, needs PFSFile (file), UIntPtr (offset), PChar (buffer) and UIntPtr (size) 
	UIntPtr (*write)(struct FSFileStruct*, UIntPtr, PChar, UIntPtr);		// Returns UIntPtr, needs PFSFile (file), UIntPtr (offset), PChar (buffer) and UIntPtr (size) 
	Int (*open)(struct FSFileStruct*);		// Returns Int, needs PFSFile (file)
	Int (*close)(struct FSFileStruct*);		// Returns Int, needs PFSFile (file)
	Boolean (*create)(struct FSFileStruct*, PChar, UInt8);		// Returns boolean, needs PFSFile (parent), PChar (name), and UInt8 (type)
	PChar (*read_directory_entry)(struct FSFileStruct*, UIntPtr);		// Returns PChar, needs PFSFile (directory) and UIntPtr (index)
	struct FSFileStruct* (*find_in_directory)(struct FSFileStruct*, PChar);		// Returns PFSFile, needs PFSFile (directory) and PChar (name)
	UIntPtr inode;
} FSFile, *PFSFile;

PFSFile FSRoot = Null;

UIntPtr FSRead(PFSFile file, UIntPtr off, PChar buf, UIntPtr size)
{
	return (!(!file || !buf || (size <= 0))) ? (file -> flags & FS_FLAG_FILE ? (file -> read ? file -> read(file, off, buf, size) : -1) : -1) : -1;
}

UIntPtr FSWrite(PFSFile file, UIntPtr off, PChar buf, UIntPtr size)
{
	return (!(!file || !buf || (size <= 0))) ? (file -> flags & FS_FLAG_FILE ? (file -> write ? file -> write(file, off, buf, size) : -1) : -1) : -1;
}

Int FSOpen(PFSFile file)
{
	return (!(!file)) ? (file -> open ? file -> open(file) : 1) : 0;		// By default, if we don't have a open func, so i think we have always permission, right?
}

Int FSClose(PFSFile file)
{
	return (!(!file)) ? (file -> close ? file -> close(file) : 1) : 0;		// By default, if we don't have a close func, so i think we don't have anything to close, right?
}

Boolean FSCreate(PFSFile parent, PChar name, UInt8 type)
{
	return (!(!parent)) ? (parent -> create ? parent -> create(parent, name, type) : False) : False;
}

PChar FSReadDirectoryEntry(PFSFile dir, UIntPtr index)
{
	return (!(!dir)) ? (dir -> read_directory_entry ? dir -> read_directory_entry(dir, index) : Null) : Null;
}

PFSFile FSFindInDirectory(PFSFile dir, PChar name)
{
	return (!(!dir || !name)) ? (dir -> find_in_directory ? dir -> find_in_directory(dir, name) : Null) : Null;
}

/* File */

PChar FileCanonicalizePath(PChar path)
{
	if (!path)
		return Null;
	if (path[0] != '/')
		return Null;
	
	PList list = ListNew();
	PChar p = StringDuplicate(path);
	PChar save = Null;
	PChar tok = StringTokenize(p, "/", &save);
	PChar final = Null;
	PChar foff = Null;
	UIntPtr fsize = 0;
	
	while (tok)
	{
		if (StringCompare(tok, "."))
			;		// Do nothing
		else if (StringCompare(tok, ".."))
		{
			if (list -> length)
				MemoryFree((UIntPtr)(ListRemove(list, list -> length - 1)));
		}
		else
		{
			PChar part = StringDuplicate(tok);
			ListAdd(list, part);
		}
		tok = StringTokenize(Null, "/", &save);
	}
	
	MemoryFree((UIntPtr)p);
	
	if (list -> length == 0)
	{
		final = (PChar)MemoryAlloc(2);
		if (final)
			StringCopy(final, "/");
	}
	else
	{
		ListForeach(i, list)
			fsize += StringGetLength((PChar)i -> data) + 1;
		
		final = foff = (PChar)MemoryAlloc(fsize);
		
		ListForeach(i, list)
		{
			StringCopy(foff, "/");
			foff++;
			StringCopy(foff, (PChar)i -> data);
			foff += StringGetLength((PChar)i -> data);
		}
	}
	
	ListFree(list);
	
	return final;
}

PChar FileJoinPath(PChar src, PChar incr)
{
	if ((!src) && (!incr))
		return Null;
	else if (!incr)
		return FileCanonicalizePath(src);
	else if (!src)
		return FileCanonicalizePath(incr);
	else if (src[0] != '/')
		return Null;
	
	UIntPtr psize = StringGetLength(src) + StringGetLength(incr) + 2;
	PChar path = Null;
	PChar poff = Null;
	
	if (incr[0] != '/' && src[StringGetLength(src)] != '/')
		psize++;
	
	path = poff = (PChar)MemoryAlloc(psize);
	
	StringCopy(poff, src);
	poff += StringGetLength(src);
	
	if (incr[0] != '/' && src[StringGetLength(src)] != '/')
		*poff++ = '/';
	
	StringCopy(poff, incr);
	
	PChar final = FileCanonicalizePath(path);
	
	MemoryFree((UIntPtr)path);
	
	return final;
}

PList FileTokenizePath(PChar path)
{
	if (!path)
		return Null;
	else if (path[0] == '\0')
		return Null;
	else if ((StringGetLength(path) == 1) && path[0] == '/')
		return ListNew();
	
	PList list = ListNew();
	
	PChar p = StringDuplicate(path);
	PChar save = Null;
	PChar pch = StringTokenize(p, "/", &save);
		
	while (pch)
	{
		if (StringCompare(pch, "."))
			;
		else if (StringCompare(pch, ".."))
		{
			if (list -> length)
				MemoryFree((UIntPtr)(ListRemove(list, list -> length - 1)));
		}
		else
		{
			PChar str = StringDuplicate(pch);
			ListAdd(list, str);
		}
		pch = StringTokenize(Null, "/", &save);
	}
	
	MemoryFree((UIntPtr)p);
	
	return list;
}

UIntPtr FileRead(PFSFile file, UIntPtr off, PChar buf, UIntPtr size)
{
	return FSRead(file, off, buf, size);
}

UIntPtr FileWrite(PFSFile file, UIntPtr off, PChar buf, UIntPtr size)
{
	return FSWrite(file, off, buf, size);
}

PFSFile FileOpen(PChar path)
{
	if (!path)
		return Null;
	if (path[0] != '/')
		return Null;
	
	PList parts = FileTokenizePath(path);
	PFSFile curr = FSRoot;
	
	if (!(FSOpen(curr)))
	{
		ListFree(parts);
		return Null;
	}
	
	ListForeach(node, parts)
	{
		curr = FSFindInDirectory(curr, (PChar)node -> data);
		
		if (curr == Null)
		{
			ListFree(parts);
			return Null;
		}
		
		if (!(FSOpen(curr)))
		{
			ListFree(parts);
			return Null;
		}
	}
	
	ListFree(parts);
	
	return curr;
}

Int FileClose(PFSFile file)
{
	return FSClose(file);
}

Boolean FileCreate(PChar path, UInt8 type, Boolean recursive)
{
	if (!path)
		return False;
	if (path[0] != '/')
		return False;
	
	PFSFile file = FileOpen(path);
	
	if (file && file -> flags == type)
	{
		FileClose(file);
		return True;
	}
	else if (file)
	{
		FileClose(file);
		return False;
	}
	
	PList parts = FileTokenizePath(path);
	PFSFile oldcurr = FSRoot;
	PFSFile curr = FSRoot;
	
	if (!(FSOpen(curr)))
	{
		ListFree(parts);
		return False;
	}
	
	ListForeach(node, parts)
	{
		oldcurr = curr;
		curr = FSFindInDirectory(curr, (PChar)node -> data);
		
		if (((curr == Null) && (node -> next == Null)) || ((curr == Null) && (recursive == True)))
		{
			FSCreate(oldcurr, (PChar)node -> data, ((node -> next == Null) ? type : FS_FLAG_DIR));
			curr = FSFindInDirectory(oldcurr, (PChar)node -> data);
		}
		
		if (curr == Null)
		{
			ListFree(parts);
			return False;
		}
		
		if (!(FSOpen(curr)))
		{
			ListFree(parts);
			return False;
		}
	}
	
	FileClose(curr);
	ListFree(parts);
	
	return True;
}

/* Mutex */

Void MutexLock(UInt8 mutex)
{
	while (mutex)
		Asm Volatile("hlt");
	mutex = 1;
}

Void MutexUnlock(UInt8 mutex)
{
	mutex = 0;
	(Void)mutex;
}

/* System calls */

typedef struct
{
	PUInt32 major;
	PUInt32 minor;
	PUInt32 build;
	PChar* codename;
	PChar* arch;
} SystemVersion, *PSystemVersion;

typedef struct
{
	UInt16 hour;
	UInt16 minute;
	UInt16 second;
} Time, *PTime;

typedef struct
{
	UInt16 month;		// TODO: Add year (we need acpi for doing this)
	UInt16 day;
} Date, *PDate;

typedef struct
{
	PFSFile file;
	Int num;
} ProcessFile, *PProcessFile;

extern PProcessInfo ProcessCurrentInfo;
extern PProcess ProcessCurrent;
extern PPageDirectory PagingCurrentDirectory;

Int ProcessAppendFile(PFSFile file)
{
	ListForeach(i, ProcessCurrentInfo -> file_list)
	{
		PProcessFile pf = (PProcessFile)i -> data;
		
		if (!pf -> file)
		{
			pf -> file = file;
			return pf -> num;
		}
	}
	
	PProcessFile pf = (PProcessFile)MemoryAlloc(sizeof(ProcessFile));
	
	if (!pf)
		return -1;
	
	pf -> file = file;
	pf -> num = ProcessCurrentInfo -> file_count++;
	
	ListAdd(ProcessCurrentInfo -> file_list, pf);
	
	return pf -> num;
}

Int SysUnimplemented(Void)
{
	return -1;
}

Int SysGetSystemVersion(PSystemVersion ver)
{
	if (ver -> major)
		*ver -> major = CHICAGO_MAJOR;
	if (ver -> minor)
		*ver -> minor = CHICAGO_MINOR;
	if (ver -> build)
		*ver -> build = CHICAGO_BUILD;
	if (ver -> codename)
		StringCopy(*ver -> codename, CHICAGO_CODENAME);
	if (ver -> arch)
		StringCopy(*ver -> arch, CHICAGO_ARCH);
	
	return 0;
}

Int SysExit(Void)
{
	ProcessExit();
	
	while (1) ;
	
	return 0;
}

UInt32 SysGetProcessID(Void)
{
	return ProcessCurrentInfo -> pid;
}

PProcessInfo SysGetProcessInfo(Void)
{
	return ProcessCurrentInfo;
}

Int SysCreateProcess(UIntPtr entry)
{
	PProcess proc = ProcessCreate(entry);
	ProcessAdd(proc);
	return proc -> id;
}

Boolean SysProcessIsRunning(UInt32 id)
{
	return ProcessGet(id) ? True : False;
}

UIntPtr SysGetHeapStart(Void)
{
	return ProcessCurrentInfo -> heap_start;
}

UIntPtr SysGetHeapCurrent(Void)
{
	return ProcessCurrentInfo -> heap_current;
}

UIntPtr SysIncrementHeap(UIntPtr incr)
{
	if (incr <= 0)
		return -1;
	
	ProcessCurrentInfo -> heap_current += incr;
	
	for (; ProcessCurrentInfo -> heap_current_aligned < ProcessCurrentInfo -> heap_current; ProcessCurrentInfo -> heap_current_aligned += 0x1000)
	{
		PagingGetPage(ProcessCurrentInfo -> heap_current_aligned, 1, PagingCurrentDirectory);
		PagingAllocFrame(PagingGetPage(ProcessCurrentInfo -> heap_current_aligned, 1, PagingCurrentDirectory), 0, 1);
	}
	
	Asm Volatile("movl %cr3, %eax; movl %eax, %cr3");
	
	return 0;
}

UIntPtr SysDecrementHeap(UIntPtr decr)
{
	if (decr <= 0)
		return -1;
	else if (ProcessCurrentInfo -> heap_current - decr < ProcessCurrentInfo -> heap_start)
		return -1;
	
	ProcessCurrentInfo -> heap_current -= decr;
	
	return 0;
}

Int SysOpenFile(PChar path)
{
	if (!path)
		return -1;
	
	PFSFile file = FileOpen(path);
	
	if (!file)
		return -1;
	
	return ProcessAppendFile(file);
}

Int SysCloseFile(Int file)
{
	if (file == 0 || file == 1 || file == 2)
		return 0;
	if (file >= ProcessCurrentInfo -> file_count)
		return -1;
	
	((PProcessFile)(ListGet(ProcessCurrentInfo -> file_list, file) -> data)) -> file = Null;
	
	return 0;
}

Int SysReadFile(Int file, PChar buf, UIntPtr size)
{
	if (file == 0)
	{
		Asm Volatile("sti");
		GetString(buf, False, size);
		Asm Volatile("cli");
		if (StringGetLength(buf) < size)
		{
			Int s = StringGetLength(buf);
			buf[s] = '\n';
			buf[s + 1] = '\0';
		}
		return StringGetLength(buf);
	}
	else if (file == 1 || file == 2)
		return -1;
	else if (file >= ProcessCurrentInfo -> file_count)
		return -1;
	
	PFSFile f = ((PProcessFile)(ListGet(ProcessCurrentInfo -> file_list, file) -> data)) -> file;
	
	if (f == Null)
		return -1;
	
	UIntPtr off = f -> offset;
	
	f -> offset += size;
	
	return FileRead(f, off, buf, size);
}

Int SysWriteFile(Int file, PChar buf, UIntPtr size)
{
	if (file == 0)
		return -1;
	if (file == 1 || file == 2)
	{
		for (UIntPtr i = 0; i < size; i++)
			ConsoleWriteCharacter(buf[i]);
		return size;
	}
	if (file >= ProcessCurrentInfo -> file_count)
		return -1;
	
	PFSFile f = ((PProcessFile)(ListGet(ProcessCurrentInfo -> file_list, file) -> data)) -> file;
	
	if (f == Null)
		return -1;
	
	UIntPtr off = f -> offset;
	
	f -> offset += size;
	
	return FileWrite(f, off, buf, size);
}

Int SysCreateFile(PChar path, UInt8 type, Boolean recursive)
{
	if (!path)
		return 0;
	
	return FileCreate(path, type, recursive);
}

UIntPtr SysSetFileOffset(Int file, UIntPtr off, UInt32 type)
{
	if (file == 0 || file == 1 || file == 2)
		return 0;
	if (file >= ProcessCurrentInfo -> file_count)
		return -1;
	
	PFSFile f = ((PProcessFile)(ListGet(ProcessCurrentInfo -> file_list, file) -> data)) -> file;
	
	if (type == 0)
		f -> offset = off;
	else if (type == 1)
		f -> offset += off;
	else if (type == 2)
		f -> offset = f -> length + off;
	else
		return -1;
	
	return f -> offset;
}

UIntPtr SysGetFileOffset(Int file)
{
	if (file == 0 || file == 1 || file == 2)
		return 0;
	if (file >= ProcessCurrentInfo -> file_count)
		return -1;
	
	return (((PProcessFile)(ListGet(ProcessCurrentInfo -> file_list, file) -> data)) -> file) -> offset;
}

Int SysSleep(UIntPtr ms)
{
	TimerSleep(ms);
	return 0;
}

Int SysGetTime(PTime time)
{
	if (!time)
		return -1;
	
	PortOutByte(0x70, 0x04);
	time -> hour = (PortInByte(0x71) & 0x80);
	
	PortOutByte(0x70, 0x02);
	time -> minute = (PortInByte(0x71) & 0x80);
	
	PortOutByte(0x70, 0x00);
	time -> second = (PortInByte(0x71) & 0x80);
	
	PortOutByte(0x70, 0x0B);
	if (!((PortInByte(0x71) & 0x80) & 0x04))
	{
		time -> hour = ((time -> hour & 0x0F) + (((time -> hour & 0x70) / 16) * 10)) | (time -> hour & 0x80);
		time -> minute = (time -> second & 0x0F) + ((time -> second / 16) * 10);
		time -> second = (time -> second & 0x0F) + ((time -> second / 16) * 10);
	}
	
	PortOutByte(0x70, 0x0B);
	if (!((PortInByte(0x71) & 0x80) & 0x02) && (time -> hour & 0x80))
		time -> hour = ((time -> hour & 0x7F) + 12) % 24;
	
	return 0;
}

Int SysGetDate(PDate date)
{
	if (!date)
		return -1;
	
	PortOutByte(0x70, 0x08);
	date -> month = (PortInByte(0x71) & 0x80);
	
	PortOutByte(0x70, 0x07);
	date -> day = (PortInByte(0x71) & 0x80);
	
	PortOutByte(0x70, 0x0B);
	if (!((PortInByte(0x71) & 0x80) & 0x04))
	{
		date -> month = (date -> month & 0x0F) + ((date -> month / 16) * 10);
		date -> day = (date -> day & 0x0F) + ((date -> day / 16) * 10);
	}
	
	return 0;
}

UIntPtr SysGetSeconds(Void)
{
	Time time;
	Date date;
	UInt16 year;
	
	SysGetTime(&time);
	SysGetDate(&date);
	
	PortOutByte(0x70, 0x09);
	year = (PortInByte(0x71) & 0x80);
	year = ((year / 16) * 10 + (year & 0x0F)) - 70;
	
	return year * 31556926 + date.month * 2629743 + date.day * 87400 + time.hour * 3600 + time.minute * 60 + time.second;
}

Void SystemCallHandler(PRegisters regs)
{
	switch (regs -> eax)
	{
		case 0x00:		// Void GetSystemVersion(PSystemVersion ver)
		{
			regs -> eax = SysGetSystemVersion((PSystemVersion)regs -> ebx);
			break;
		}
		case 0x01:		// Void Exit(Void)
		{
			regs -> eax = SysExit();
			break;
		}
		case 0x02:		// UInt32 GetProcessID(Void)
		{
			regs -> eax = SysGetProcessID();
			break;
		}
		case 0x03:		// PProcessInfo GetProcessInfo(Void)
		{
			regs -> eax = (UIntPtr)SysGetProcessInfo();
			break;
		}
		case 0x04:		// Void CreateProcess(UIntPtr entry)
		{
			regs -> eax = SysCreateProcess(regs -> ebx);
			break;
		}
		case 0x05:		// Boolean ProcessIsRunning(UInt32 id)
		{
			regs -> eax = SysProcessIsRunning(regs -> ebx);
			break;
		}
		case 0x06:		// UIntPtr GetHeapStart(Void)
		{
			regs -> eax = SysGetHeapStart();
			break;
		}
		case 0x07:		// UIntPtr GetHeapCurrent(Void)
		{
			regs -> eax = SysGetHeapCurrent();
			break;
		}
		case 0x08:		// UIntPtr IncrementHeap(UIntPtr incr)
		{
			regs -> eax = SysIncrementHeap(regs -> ebx);
			break;
		}
		case 0x09:		// UIntPtr DecrementHeap(UIntPtr decr)
		{
			regs -> eax = SysDecrementHeap(regs -> ebx);
			break;
		}
		case 0x0A:		// Int OpenFile(PChar path)
		{
			regs -> eax = SysOpenFile((PChar)regs -> ebx);
			break;
		}
		case 0x0B:		// Int CloseFile(Int file)
		{
			regs -> eax = SysCloseFile(regs -> ebx);
			break;
		}
		case 0x0C:		// Int ReadFile(Int file, PChar buf, UIntPtr size)
		{
			regs -> eax = SysReadFile(regs -> ebx, (PChar)regs -> ecx, regs -> edx);
			break;
		}
		case 0x0D:		// Int WriteFile(Int file, PChar buf, UIntPtr size)
		{
			regs -> eax = SysWriteFile(regs -> ebx, (PChar)regs -> ecx, regs -> edx);
			break;
		}
		case 0x0E:		// Int CreateFile(PChar path, UInt8 type, Boolean recursive)
		{
			regs -> eax = SysCreateFile((PChar)regs -> ebx, regs -> ecx, regs -> edx);
			break;
		}
		case 0x0F:		// UIntPtr SetFileOffset(Int file, UIntPtr off, UInt32 type)
		{
			regs -> eax = SysSetFileOffset(regs -> ebx, regs -> ecx, regs -> edx);
			break;
		}
		case 0x10:		// UIntPtr GetFileOffset(Int file)
		{
			regs -> eax = SysGetFileOffset(regs -> ebx);
			break;
		}
		case 0x11:		// Void Sleep(UIntPtr ms)
		{
			regs -> eax = SysSleep(regs -> ebx);
			break;
		}
		case 0x12:		// Int GetTime(PTime time)
		{
			regs -> eax = SysGetTime((PTime)regs -> ebx);
			break;
		}
		case 0x13:		// Int GetDate(PDate date)
		{
			regs -> eax = SysGetDate((PDate)regs -> ebx);
			break;
		}
		case 0x14:		// UIntPtr GetSeconds(Void)
		{
			regs -> eax = SysGetSeconds();
			break;
		}
		default:		// Unimplemented
		{
			regs -> eax = SysUnimplemented();
			break;
		}
	}
}

/* Kernel */

Void KernelMain(PBootData bootdata)
{
	/* Initialize console */
	
	ConsoleClear();
	Message(MESSAGE_LEVEL_INFO, "Console initialized\n");
	
	/* We're booting with GRUB, right? */
	
	if (bootdata -> mbmagic != 0x2BADB002)
		Panic("Sorry, CHicago got an fatal error and halted.\nThis fault happened in KernelMain() because the multiboot magic is wrong.");		// No, but we need some informations from GRUB, so halt
	
	/* Architecture-specific initialization */
	
	ArchInit(bootdata -> mbheader);
	
	/* Halt, as the architecture-specific code will initialize tasking, and go to KernelMainLate */
	
	while (1) ;
}

Void KernelMainLate(Void)
{	
	/* Init system calls */
	
	IDTRegisterInterruptHandler(0x3F, SystemCallHandler);
	Message(MESSAGE_LEVEL_INFO, "System calls initialized\n");
	
	/* Halt, we don't have anything to do */
	
	while (1) ;
}
